"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SpreadMatcher = void 0;
const lodash_1 = require("lodash");
const bignumber_1 = require("@ethersproject/bignumber");
var SpreadMatcher;
(function (SpreadMatcher) {
    SpreadMatcher.spreadMatcher = (originalData) => {
        const clonedData = lodash_1.cloneDeep(originalData);
        const data = {
            ...clonedData,
            entityGroups: [
                clonedData.entityGroups[0].map((e) => ({ ...e, routes: [] })),
                clonedData.entityGroups[1].map((e) => ({ ...e, routes: [] }))
            ]
        };
        const entityMap = [
            new Map(data.entityGroups[0].map((e) => [e, e])),
            new Map(data.entityGroups[1].map((e) => [e, e]))
        ];
        const matchRoundResults = [];
        let i = 0;
        while ((i += 1)) {
            if (i > 10000) {
                throw new Error(`
                  Matching executed more than 10_000 - it may mean, that something broke
                  data: ${JSON.stringify(clonedData)}
                  matches: not displayed, because of potential length
                `);
            }
            if (data.entityGroups[0].length === 0 || data.entityGroups[1].length === 0) {
                break;
            }
            const roundInput = getRoundInput(data);
            if (roundInput.routes.length === 0) {
                break;
            }
            const roundMatches = matchRound(roundInput);
            roundMatches.forEach((match) => {
                const entity1 = entityMap[0].get(match.entities[0]);
                const entity2 = entityMap[1].get(match.entities[1]);
                entity1.volume = entity1.volume.sub(match.volume);
                entity2.volume = entity2.volume.sub(match.volume);
            });
            matchRoundResults.push(roundMatches);
            data.entityGroups = roundInput.entityGroups;
        }
        data.entityGroups[0].forEach((e) => delete e.routes);
        data.entityGroups[1].forEach((e) => delete e.routes);
        matchRoundResults.flat().forEach((m) => m.entities.forEach((e) => delete e.routes));
        return {
            matches: sumMatches(matchRoundResults.flat()).map(omitMatchVolumes),
            roundMatches: matchRoundResults,
            leftoverEntities: [
                data.entityGroups[0].filter((e) => e.volume.gt(0)),
                data.entityGroups[1].filter((e) => e.volume.gt(0))
            ]
        };
    };
    const omitMatchVolumes = (match) => {
        return {
            ...match,
            entities: match.entities.map((e) => lodash_1.omit(e, 'volume'))
        };
    };
    const sumMatches = (matches) => {
        const matchMap = new Map();
        matches.forEach((match) => {
            const entity1Entry = matchMap.get(match.entities[0]) ?? new Map();
            matchMap.set(match.entities[0], entity1Entry);
            const entity2Entry = entity1Entry.get(match.entities[1]) ?? bignumber_1.BigNumber.from(0);
            entity1Entry.set(match.entities[1], entity2Entry.add(match.volume));
        });
        const entity1Entries = Array.from(matchMap.entries());
        const result = entity1Entries.map(([entity1, entity1Entry]) => {
            const entity2Entries = Array.from(entity1Entry.entries());
            return entity2Entries.map(([entity2, volume]) => ({
                entities: [entity1, entity2],
                volume
            }));
        });
        return result.flat();
    };
    const computeSafeDistribution = (entities1, entities2) => {
        const result = bigNumMinBy([...entities1, ...entities2], (e) => e.volume.div(e.routes.length));
        return {
            entity: result,
            volume: result.volume.div(result.routes.length)
        };
    };
    const matchRound = (input) => {
        const toDistribute = computeSafeDistribution(input.entityWithRoutes[0], input.entityWithRoutes[1]);
        if (toDistribute.volume.eq(0)) {
            const competingEntities = toDistribute.entity.routes;
            const isEntityFirst = input.entityWithRoutes[0].some((e) => e === toDistribute.entity);
            const entitiesToSatisfy = competingEntities.slice(0, toDistribute.entity.volume.toNumber());
            const result = entitiesToSatisfy.map((connectedEntity) => ({
                entities: (isEntityFirst
                    ? [toDistribute.entity, connectedEntity]
                    : [connectedEntity, toDistribute.entity]),
                volume: bignumber_1.BigNumber.from(1)
            }));
            return result;
        }
        return input.routes.map((route) => ({
            entities: route,
            volume: toDistribute.volume
        }));
    };
    const getRoundInput = (data) => {
        const entities = [
            prepareEntitiesForRoundInput(data.entityGroups[0]),
            prepareEntitiesForRoundInput(data.entityGroups[1])
        ];
        const entity1Group = getFirstUsableGroup(data.groupPriority, entities[0].grouped);
        const routes = entity1Group.flatMap((entity1GroupEntry) => {
            const preferredEntity2Group = getFirstUsableGroup(entity1GroupEntry.groupPriority, entities[1].grouped);
            return cartesian(entity1GroupEntry.entities, preferredEntity2Group.flatMap((g) => g.entities));
        });
        routes.forEach(([entity1, entity2]) => {
            entity1.routes.push(entity2);
            entity2.routes.push(entity1);
        });
        return {
            routes,
            entityWithRoutes: [
                entities[0].flat.filter((entity) => entity.routes.length > 0),
                entities[1].flat.filter((entity) => entity.routes.length > 0)
            ],
            entityGroups: [entities[0].flat, entities[1].flat]
        };
    };
    const prepareEntitiesForRoundInput = (entities) => {
        const grouped = {};
        const flat = [];
        const entitiesLength = entities.length;
        for (let i = 0; i < entitiesLength; i += 1) {
            const entity = entities[i];
            if (entity.volume.eq(0)) {
                continue;
            }
            entity.routes = [];
            if (!(entity.id in grouped)) {
                grouped[entity.id] = [];
            }
            grouped[entity.id].push(entity);
            flat.push(entity);
        }
        return {
            grouped,
            flat
        };
    };
    const getFirstUsableGroup = (groups, entities) => {
        const groupsWithEntities = groups.map((group) => {
            const groupWithEntities = group.map((entry) => ({
                ...entry,
                entities: entities[entry.id] ?? []
            }));
            return groupWithEntities;
        });
        const firstAvailableEntry = groupsWithEntities.find((group) => group.some((g) => g.entities.length > 0));
        return firstAvailableEntry ?? [];
    };
    const bigNumMinBy = (collection, cb) => {
        let minResult = cb(collection[0]);
        let minItem = collection[0];
        for (let i = 0; i < collection.length; i += 1) {
            const currentResult = cb(collection[i]);
            if (currentResult.lt(minResult)) {
                minItem = collection[i];
                minResult = currentResult;
            }
        }
        return minItem;
    };
    const cartesian = (arr1, arr2) => {
        return arr1.flatMap((val1) => arr2.map((val2) => [val1, val2]));
    };
})(SpreadMatcher = exports.SpreadMatcher || (exports.SpreadMatcher = {}));
//# sourceMappingURL=spreadMatcher.js.map